#!/usr/bin/perl -w

# @Java installation/configuration script
# Author: Federico Pedemonte <pedemonte@linux.it>

use strict 'vars';

# -------- Don't change this unless you know exactly what you are doing -------
my $RPM_MODE = 0;
my $LIB_DIR;
my $ATJAVAC_BIN = '/usr/bin/atjavac';
# ----------------------------------------------------------------------------

my $TMP_DIR;	

my $AJC;
my $JAVA_HOME;

my $conf_file		= "/etc/atjava.conf";

# filename of the original rt.jar 
my $RT_JAR			= "/jre/lib/rt.jar";				

# filename of the original rj.jar backup 
my $BACKUP_RT_JAR	= "/jre/lib/rt.jar_atjava";			

# filename of the jar command    
my $JAR 			= "/bin/jar";						

my $BCEL;
my $JAKARTA_REGEXP;
my $CLASSPATH;

my $DEST_DIR	= "/usr/local/atjava/";				

my $uninstall = 0;
my $readconf = undef;
my $keepconf = 0;	#not really used...

# ----------------------------------------------------------------------------
# finds the location of an executable in $PATH 
# usage : which (bin_name)
# ----------------------------------------------------------------------------
sub which {
	my $bin_name = shift;
	my @paths;
	my $path;
	my $full_name;
	
	if (substr($bin_name, 0, 1) eq '/') {
		# absolute path
		if ((-f $bin_name) && (-x $bin_name)) {
			return $bin_name;
		}
	} else {
		# relative path
		foreach $path (split(':', $ENV{'PATH'})) {
			$full_name = $path . '/' . $bin_name;
			if ((-f $full_name) && (-x $full_name)) {
				return $full_name;
			}
		}
	}
	
	# not found
	return;
}

# ----------------------------------------------------------------------------
# finds the location of a jar library in the CLASSPATH specified directories
# usage : which_jar (jar_name)
# ----------------------------------------------------------------------------
sub which_jar {
	my $jar_name = shift;
	my @paths;
	my $path;
	my $full_name;
	
	foreach $path (split(':', $ENV{'CLASSPATH'})) {
		$full_name = $path;
		
		#TODO: write a better check condition
		if ($full_name =~ $jar_name) {
			return $full_name;
		}	
	}
	
	# not found
	return;
}

# ----------------------------------------------------------------------------
# asks for user input
# usage : ask_user (question, default answer)
# ----------------------------------------------------------------------------
sub ask_user {
	my ($askString,$defaultValue) = @_;
	my $input;
	
	if ($defaultValue) {
		print $askString, "[", $defaultValue, "]: ";
	} else {
		print $askString, ": ";
	}
	
	#$| = 1;            # flush
	$input = <STDIN>;	# reads the input from STDIN (presumably the keyboard)
	
	chomp $input;
	
	if (defined $defaultValue) {
		return $input ? $input : $defaultValue;    # return the user input (if exists)
	} else {
		return $input;
	}
}

# ----------------------------------------------------------------------------
# check whether atjava is already installed by checking
# if a copy of the original rt.jar exists with the expected name
# usage: is_already_installed (path_to_check)
# ----------------------------------------------------------------------------
sub is_already_installed {
	my $java_home = shift;
	return (-f ($java_home.$BACKUP_RT_JAR));
}

# ----------------------------------------------------------------------------
# applies the patches to the original Java rt.jar
# and keeps a copy of the original file
# ----------------------------------------------------------------------------
sub patch_rt_jar { 
	# If libdir doesn't starts with /, then adds the full path of the 
	# current working dir in front of it
	my $cur_path;
	
	if (! ($LIB_DIR =~ /^\//)) {
		$cur_path = `pwd`;
		chomp $cur_path;
		$LIB_DIR = $cur_path."/".$LIB_DIR;
	}
	
	# creates a temporary directory
	$TMP_DIR = `mktemp -t -d "at_java.XXXX"` || die "ERRROR: Unable to create temporary directory! \@Java installation aborted\n";
	chomp $TMP_DIR;
	chdir "$TMP_DIR";
	
	# makes a copy of rt.jar
	system ("cp -p $JAVA_HOME$RT_JAR $JAVA_HOME$BACKUP_RT_JAR");
	
	# create a new rt.jar
	system ("cp -p $JAVA_HOME$RT_JAR .");	
	
	# --- TO BE REMOVED  ------
	system ("$JAVA_HOME$JAR xf $BCEL");
	# --- END OF - TO BE REMOVED  ------
	
    system ("$JAVA_HOME$JAR xf $LIB_DIR/lib/atjava.jar");
	system ("$JAVA_HOME$JAR xf $LIB_DIR/lib/atjavac.jar");
	
	system ("$JAVA_HOME$JAR xf rt.jar java/lang/reflect/Method.class java/lang/reflect/Constructor.class java/lang/reflect/Field.class");
    
    system ("export CLASSPATH=$CLASSPATH; $AJC -inpath java -aspectpath atjava/lang/annotation/ -d .");
	
    system ("$JAVA_HOME$JAR uf rt.jar atjavac atjava java org");
	
	# replace the original rt.jar with the new one
	system ("mv rt.jar $JAVA_HOME$RT_JAR");
	
	# remove temporary directory
	if ((defined $TMP_DIR) && ($TMP_DIR ne "") && ($TMP_DIR ne "/") && ($TMP_DIR ne ".") && ($TMP_DIR ne "..")) {
		system ("rm -Rf $TMP_DIR");
	}
}

# ----------------------------------------------------------------------------
# returns full path of the passed jar archive
# first check in the CLASSPATH for valid location
# ----------------------------------------------------------------------------
sub get_jar_location {
	my $jar_name = shift ;
	my $full_path;
	
	$full_path = which_jar($jar_name);
	
	print "\n";
	print "We need to locate the $jar_name Java library\n";
	
	if (defined $full_path) {
		print "A valid file has already been found in your CLASSPATH environment variable\n";
		print "Please confirm it or provide the full name of the file you'd like to use\n";
	} else {
		print "Your environment variable CLASSPATH doesn't contain any reference to that library\n";
		print "Please provide a full path name for it\n";
	}
	
	while (1) {
		$full_path = ask_user ("$jar_name location? (or x to abort) ", $full_path);
		
		if ($full_path eq 'x') { 
			return;
		}
		
		if (-f $full_path) {
			return $full_path;
		} else {
			print "\nFile doesn't exist. Please specify a correct file name\n";		
		}
	}
}

# ----------------------------------------------------------------------------
# returns full path of the passed bin file
# first check in the $PATH variable for valid location
# ----------------------------------------------------------------------------
sub get_bin_location {
	my $bin_name = shift ;
	my $full_path;
	
	$full_path = which($bin_name);
	
	print "\n";
	print "We need to locate the $bin_name executable\n";
	
	if (defined $full_path) {
		print "A valid file has already been found in your PATH environment variable\n";
		print "Please confirm it or provide the full name of the file you'd like to use\n";
	} else {
		print "Your PATH doesn't contain any reference to that executable\n";
		print "Please provide a full path name for it\n";
	}
	
	while (1) {
		$full_path = ask_user ("$bin_name location? (or x to abort) ", $full_path);
		
		if (defined ($full_path)) {
			if ($full_path eq 'x') { 
				return;
			}
			
			if ((-f $full_path) && (-x $full_path)) {
				return $full_path;
			} else {
				print "\nFile doesn't exist or is not executable. Please specify a correct file name\n";		
			}
		}
	}
}

# ----------------------------------------------------------------------------
# checks if the passed path is a valid java location
# ----------------------------------------------------------------------------
sub is_valid_java_home {
	my $java_home = shift ;
	my $valid;
	
	return NULL unless defined $java_home;
	
	if ((-f $java_home.$JAR) && (-x $java_home.$JAR)) {
		$valid = 1;
	} else {
		$valid = 0;
	}
	
	return $valid;
}

# ----------------------------------------------------------------------------
# returns JAVA_HOME environment variable if set
# otherwise ask user and returns the answer
# ----------------------------------------------------------------------------
sub get_java_home {
	my $java_home;
	
	print "\n";
	
	if (defined $ENV{'JAVA_HOME'}) {
		$java_home = $ENV{'JAVA_HOME'};
		if (is_valid_java_home ($java_home)) {
			print "The environment variable JAVA_HOME seems to have a valid value.\nPlease press ENTER to confirm it or specify a new one\n";	
		} else {
			print "The environment variable JAVA_HOME seems to have an invalid value. \nPlease specify a new one\n";
		}
	} else {
		print "It seems that the environment variable JAVA_HOME is not set\n";
		print "Please specify the path to your local java installation\n";
	}
	
	while (1) {
		print "\n";
		
		$java_home = ask_user ("Java location (JAVA_HOME)? (or x to abort)", $java_home);	
		
		if ($java_home eq 'x') {
			return;
		}
		
		if (!is_valid_java_home($java_home)) {
			print "$java_home seems to be an invalid JAVA_HOME location\n";	
		} else {
			return $java_home;	
		}
	}
	return $java_home;
}

# ----------------------------------------------------------------------------
# INSTALL procedure
# ----------------------------------------------------------------------------
sub install () {
	print "\@Java Installation Procedure\n";
	
	#checks if a configuration file already exists
	if (-f $conf_file) {
		print "A configuration file for \@Java ($conf_file) is already present\n";
		print "If you want to change your configuration or patch another SDK, you have to uninstall \@Java first\n";
		exit ();
	}
	
	# sets java location
	$JAVA_HOME = get_java_home ();
	if (!defined $JAVA_HOME) {
		print "\nAborting. Bye\n\n";
		exit ();		
	}
	
	# checks if @Java is already installed in the provided SDK	
	if (is_already_installed($JAVA_HOME)) {
		die "ERROR: It seems that atjava is already installed for that SDK. Aborting\n";	
	}
	
	# sets BCEL location
	$BCEL = get_jar_location ("bcel");
	if (!defined $BCEL) {
		print "\nAborting. Bye\n\n";
		exit ();	
	}

    # TODO: Remove all the Jakarta Regexp reference
	# sets Jakarta Regexp location
	# $JAKARTA_REGEXP = get_jar_location ("jakarta-regexp");
	# if (!defined $JAKARTA_REGEXP) {
	# 	print "\nAborting. Bye\n\n";
	# 	exit ();
	# }
    $JAKARTA_REGEXP="";
	
	# sets AspectJ binary location
	$AJC = get_bin_location("ajc");
	if (!defined $AJC) {
		print "\nAborting. Bye\n\n";
		exit ();		
	}
	
	# if ajc path was not set, we assume that neither Aspectrt.jar was in the CLASSPATH,
	# so we add a possible path to it
	$AJC =~ m/(.*)\/bin\/ajc/g;	
	$CLASSPATH = $CLASSPATH.":".$1."/lib/aspectjrt.jar";	
	
	# apply patches to rt.jar	
	patch_rt_jar ();
	
	# if not in RPM mode, we need to phisically copy 
    # the libraries, the executables and the libraries 
	# to the destination directories
	if (!$RPM_MODE) {
		install_exec ();
	}

	patch_scripts ();
    # makes symlink to executables
    system ("ln -fs $DEST_DIR/bin/atjavac $ATJAVAC_BIN");
	
	# stores the information about the system collected into a configuration file
	# in order to properly disinstall @Java
	write_config ();

	print "\@Java successfully installed\n";
}

# ----------------------------------------------------------------------------
# install_exec ()
# install the scripts/libraries/documentation in the destination dirs
# ----------------------------------------------------------------------------
sub install_exec () {
	#copy all the files to the destination
	system ("mkdir $DEST_DIR");
	system ("cp -Rp $LIB_DIR/* $DEST_DIR ");
	#TODO:change ownership and permission
}

# ----------------------------------------------------------------------------
# patch_scripts ()
# patch atjavac script and writes CLASSPATH inside
# ----------------------------------------------------------------------------
sub patch_scripts () {	
	my $file_content;	
	my $path_string;

	# sets additional CLASSPATH entries in atjavac script
	open FILE, "<$DEST_DIR/bin/atjavac";
	$file_content = do {local $/; <FILE>};
	close FILE;

	$path_string .= "CLASSPATH=\$CLASSPATH:$DEST_DIR/lib/atjava.jar\n";
	$path_string .= "CLASSPATH=\$CLASSPATH:$DEST_DIR/lib/atjavac.jar\n";

	$file_content =~ s/#CLASSPATH#/$path_string/;
	
	open FILE, ">$DEST_DIR/bin/atjavac";
	print FILE $file_content;
	close FILE;

	# makes atjavac an executable file
	system ("chmod a+x $DEST_DIR/bin/atjavac");
}

# ----------------------------------------------------------------------------
# RECONFIGURE procedure 
# called when we have an existing ajava configuration file
# ----------------------------------------------------------------------------
sub reconfigure () {
	if (read_config ($readconf)) {
		#checks if all the needed variables are set
		return unless $BCEL and $JAVA_HOME and $AJC and $JAKARTA_REGEXP;
		patch_rt_jar ();
	} 
}

# ----------------------------------------------------------------------------
# UNINSTALL procedure 
# ----------------------------------------------------------------------------
sub uninstall () {
	print "\@Java disinstallation Procedure\n";
		
	if (read_config($conf_file)) {
        # remove the symlinks to executable files
        system ("rm $ATJAVAC_BIN");
	
		# removes the current rt.jar that is supposed to be the @Java one
		system ("rm $JAVA_HOME$RT_JAR");
	
		# restores the originar rt.jar
		system ("mv $JAVA_HOME$BACKUP_RT_JAR $JAVA_HOME$RT_JAR");
	
		# deletes configuration file from system
		system ("rm $conf_file");
	} 
	
	print "\@Java successfully uninstalled\n";
}

# ----------------------------------------------------------------------------
# Writes the @Java configuration
# ----------------------------------------------------------------------------
sub write_config () {
	my @vars = qw (BCEL JAVA_HOME AJC JAKARTA_REGEXP);
	my $value;
	
	open (FILE, "> $conf_file") || die "Fatal ERROR: cannot write file $conf_file. ABORTING\n";
	
	print "Writing configuration file...\n";
	
	foreach my $v (@vars) {
        eval "\$value = \$$v";
        print FILE "\$$v=\"$value\"\n";
	}
	
	print "DONE writing configuration file\n";
	
	close (FILE);
}

# ----------------------------------------------------------------------------
# Reads the @Java configuration from the specified file
# returns 0 if the file doesn't exist
#         1 otherwise
# ----------------------------------------------------------------------------
sub read_config () {
	my $file = shift;
	my $stmnt;
	
	open (FILE, "< $file") || return 0;
	
	while ($stmnt = <FILE>) {
        if ($stmnt =~ /.*(\$BCEL|\$JAVA_HOME|\$AJC|\$JAKARTA_REGEXP)\=(.*)/) {
			eval $stmnt;
        }
	}
	
	close FILE;
	return 1;
}

# ----------------------------------------------------------------------------
# Main Block
# ----------------------------------------------------------------------------

# Parse command line arguments
for (my $arg=0; $arg < @ARGV; $arg ++) {
	if ($ARGV[$arg] =~ /^-dest_dir/) {
		(undef, $DEST_DIR) = split'=',$ARGV[$arg];
		#TODO: se DEST_DIR e' vuoto uscire con errore !
		next;
	} 
	if ($ARGV[$arg] =~ /^-lib_dir/) {
		(undef, $LIB_DIR) = split'=',$ARGV[$arg];
		#TODO: se LIB_DIR e' vuoto uscire con errore !
		next;
	} 
	if ($ARGV[$arg] eq "-u") {
		$uninstall = 1;		
		next;
	} 
	if ($ARGV[$arg] eq "-keepconf") {
		$keepconf = 1;	
		next;
	} 
	if ($ARGV[$arg]  =~ /^-readconf/) {
	 	(undef, $readconf) = split'=',$ARGV[$arg];	
		#TODO: se readconf e' vuoto uscire con errore !
		next;
	} 
	die "unknown option: $ARGV[$arg]. Aborting\n"
}

# Set defaults values
if (!defined ($LIB_DIR)) {
	# locates from where the script was executed, so where the
	# @Java files are
	$LIB_DIR = substr($0, 0, rindex($0, '/') + 1)."/install/atjava";
	chomp $LIB_DIR;
}

$CLASSPATH=$ENV{'CLASSPATH'};

if ($readconf) {
	reconfigure ();
} else {
	# Calls install/uninstall subroutines
	if ($uninstall) {
		uninstall();	
	} else {
		install();
	}
}
